import math
import shutil
import urllib.request
from pathlib import Path
from earthscii.etopo_loader import load_etopo_as_sphere_points
from earthscii.utils import log


TILE_DIR = Path("./tiles")
TILE_DIR.mkdir(parents=True, exist_ok=True)
PARTIAL_SUFFIX = ".partial"


def estimate_fov_from_screen(forward_vec, zoom, screen_width, screen_height,
                             aspect_ratio=0.5):
    """Estimate the angular field of view in degrees covered by the screen."""
    EARTH_RADIUS = 6371000  # meters
    screen_width_m = screen_width / zoom
    screen_height_m = screen_height / (zoom * aspect_ratio)

    # Convert visible width on screen (in meters) into angle at Earth's surface
    angular_width_deg = math.degrees(screen_width_m / EARTH_RADIUS)
    angular_height_deg = math.degrees(screen_height_m / EARTH_RADIUS)
    log(f"[\033[92mDEBUG\033[0m] FOV: {angular_width_deg:.2f}° x {angular_height_deg:.2f}°")
    return angular_width_deg, angular_height_deg


def etopo2022_filename(lat, lon):
    ns = 'N' if lat >= 0 else 'S'
    ew = 'E' if lon >= 0 else 'W'
    return f"ETOPO_2022_v1_15s_{ns}{abs(lat):02d}{ew}{abs(lon):03d}_surface.tif"


def download_etopo2022_tile(lat, lon):
    fname = etopo2022_filename(lat, lon)
    local_path = TILE_DIR / fname
    partial_path = local_path.with_suffix(local_path.suffix + PARTIAL_SUFFIX)

    if local_path.exists():
        return str(local_path)
    if partial_path.exists():
        print(f"Removing incomplete download: {partial_path}")
        partial_path.unlink()

    base_url = "https://www.ngdc.noaa.gov/mgg/global/relief/ETOPO2022/data/15s/15s_surface_elev_gtif/"
    url = base_url + fname
    print(f"Downloading {fname}...")
    try:
        with urllib.request.urlopen(url) as response:
            with open(partial_path, "wb") as out_file:
                shutil.copyfileobj(response, out_file)
            partial_path.rename(local_path)
            return str(local_path)
    except Exception as e:
        print(f"Error downloading {url}: {e}")
        if partial_path.exists():
            partial_path.unlink()
    return None


def latlon_from_vector(v):
    x, y, z = v
    lat = math.degrees(math.asin(z / math.sqrt(x*x + y*y + z*z)))
    lon = math.degrees(math.atan2(y, x))
    return lat, lon


def vector_from_latlon(lat, lon):
    """Converts geographic coordinates to a 3D unit vector."""
    lat_rad = math.radians(lat)
    lon_rad = math.radians(lon)
    x = math.cos(lat_rad) * math.cos(lon_rad)
    y = math.cos(lat_rad) * math.sin(lon_rad)
    z = math.sin(lat_rad)
    return x, y, z


def down_to_15(x): return (x // 15) * 15

def up_to_15(x): return ((x+14) // 15) * 15

def get_visible_tile_coords(forward_vec, fov_lat, fov_lon, padding):
    lat_center, lon_center = latlon_from_vector(forward_vec)

    lat_min = int(lat_center - fov_lat // 2) - padding
    lat_max = int(lat_center + fov_lat // 2) + padding
    lon_min = int(lon_center - fov_lon // 2) - padding
    lon_max = int(lon_center + fov_lon // 2) + padding

    lat_min = max(-85, lat_min)
    lat_max = min(85, lat_max)

    lat_min_aligned = down_to_15(lat_min)
    lat_max_aligned = up_to_15(lat_max)
    lon_min_aligned = down_to_15(lon_min)
    lon_max_aligned = up_to_15(lon_max)

    return [
        (lat, lon)
        for lat in range(lat_min_aligned, lat_max_aligned + 1, 15)
        for lon in range(lon_min_aligned, lon_max_aligned + 1, 15)
    ]


def load_visible_globe_points(forward_vec, zoom, screen_width, screen_height,
                              aspect_ratio=0.5):
    fov_horiz, fov_vert = estimate_fov_from_screen(forward_vec, zoom,
                                                   screen_width, screen_height,
                                                   aspect_ratio)
    tile_coords = get_visible_tile_coords(forward_vec, fov_vert, fov_horiz,
                                          padding=1)
    paths = []
    for lat, lon in tile_coords:
        path = download_etopo2022_tile(lat, lon)
        log(f"[\033[32mINFO\033[0m] Downloading tile: lat={lat}, lon={lon}")
        if path:
            paths.append(path)

    stride = compute_lod_stride(zoom)
    all_points = []
    for path in paths:
        try:
            points = load_etopo_as_sphere_points(path, stride=stride)
            if points is None:
                log(f"[\033[31mERROR\033[0m] No points returned from: {path}")
                continue
        except Exception as e:
            log(f"[\033[31mEXCEPTION\033[0m] Failed to load {path}: {e}")
            continue
        all_points.extend(points)

    log(f"[\033[92mDEBUG\033[0m] Looking for tiles near: {latlon_from_vector(forward_vec)}")
    log(f"[\033[92mDEBUG\033[0m] Tile coords to load: {tile_coords}")
    log(f"[\033[32mINFO\033[0m] Loaded {len(paths)} tile(s)")
    log(f"[\033[32mINFO\033[0m] Loaded {len(all_points)} points")

    if not all_points:
        log("[\033[91mFATAL\033[0m] No points to render!")
        raise RuntimeError("No terrain data loaded.")

    return all_points


def compute_lod_stride(zoom):
    log(f"[\033[92mDEBUG\033[0m] stride in use with zoom: {zoom}")
    if zoom > 3.0:
        return 2
    elif zoom > 2.0:
        return 4
    elif zoom > 1.0:
        return 8
    elif zoom > 0.5:
        return 16
    else:
        return 32



"""Display a 3D map in a terminal window."""
import curses
import time
import argparse
import numpy as np
import rasterio
import datetime
from rasterio.transform import xy
from earthscii.projection import project_map
from earthscii.renderer import render_map
from earthscii.map_loader import load_dem_as_points
from earthscii.globe_projection import project_globe
from earthscii.globe_tile_manager import load_visible_globe_points
from earthscii.globe_tile_manager import vector_from_latlon
from earthscii.utils import log
from importlib.resources import files


def main_wrapper():
    """Entry point for pip-installed script"""
    # Parse command-line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("tile", nargs="?", help="Path to a local .tif DEM tile")
    parser.add_argument("--globe", action="store_true", help="Enable global view (janky for now)")
    parser.add_argument("--lat", type=float, help="Initial latitude—global view only")
    parser.add_argument("--lon", type=float, help="Initial longitude—global view only")
    parser.add_argument( "--aspect", type=float, default=None, help="Override aspect ratio (default: 0.5), because fonts are taller than they are wide")
    parser.add_argument("--demo", action="store_true", help="Run with a bundled demo tile")
    parser.add_argument("--tilewalk", action="store_true", help="Explore ETOPO tiles one at a time (requires internet)")

    args = parser.parse_args()

    if not any(vars(args).values()):
        parser.print_help()
        return

    curses.wrapper(lambda stdscr: main(stdscr, args))

def tilewalk_mode(stdscr, args):
    """Load and display tiles individually, allowing movemeent between them."""
    log("[\033[33mWARN\033[0m] tilewalk mode not yet implemented")
    time.sleep(2)
    raise KeyboardInterrupt()


def globe_mode(stdscr, args, angle_x, angle_y, zoom, width, height, aspect_ratio):
    """Run with a globe view"""
    if args.lat is not None and args.lon is not None:
        forward_vec = np.array(vector_from_latlon(args.lat, args.lon))

        # Derive angles from vector
        angle_x = int(np.degrees(np.arcsin(forward_vec[1])))  # pitch from y
        angle_y = int(np.degrees(np.arctan2( forward_vec[2], forward_vec[0])))  # yaw from z/x
    else:
        forward_vec = forward_from_angles(angle_x, angle_y)

        log(f"[\033[92mDEBUG\033[0m] forward_vec = {forward_vec}")
        log(f"[\033[92mDEBUG\033[0m] zoom = {zoom}, screen = {width}x{height}")

    globe_points = load_visible_globe_points(
        forward_vec, zoom, width, height, aspect_ratio
    )

    return forward_vec, angle_x, angle_y, globe_points


def single_tile_mode(stdscr, args):
    """Run with a single tile"""
    log("[\033[33mWARN\033[0m] single tile mode not yet implemented")
    time.sleep(2)
    raise KeyboardInterrupt()


def main(stdscr, args):
    log("[\033[32mINFO\033[0m] New session started\n------------")

    init_curses(stdscr)

    angle_x = 0
    angle_y = 90
    angle_z = 0
    zoom = 0.8
    height, width = stdscr.getmaxyx()
    aspect_ratio = args.aspect if args.aspect is not None else min(height / width, 0.4)
    is_global = False

    if args.tilewalk:
        tilewalk_mode(stdscr, args)
    elif args.globe:
        forward_vec, angle_x, angle_y, globe_points = globe_mode(
            stdscr, args, angle_x, angle_y, zoom, width, height, aspect_ratio
        )
        is_global = True
    else:
        if args.demo:
            # Load the bundled tile using importlib.resources
            demo_path = files("earthscii").joinpath("data/n37_w123_1arc_v3.tif")
            path = str(demo_path)
        elif args.tile:
            path = args.tile
        else:
            raise ValueError("You must provide a tile path or use --demo.")

        map_data, transform = load_dem_as_points(path)

    offset_x, offset_y = width // 2, height // 2
    prev_state = None

    buffer = curses.newwin(height, width, 0, 0)

    while True:
        try:
            key = stdscr.getch()

            angle_x, angle_y, angle_z, zoom, offset_x, offset_y, changed = handle_keys(
                key, angle_x, angle_y, angle_z, zoom, offset_x, offset_y
            )

            if changed:
                stdscr.refresh()
                if is_global:
                    forward_vec = forward_from_angles(angle_x, angle_y)
                    globe_points = load_visible_globe_points(
                        forward_vec, zoom, width, height, aspect_ratio
                    )


            state = (angle_x, angle_y, angle_z, zoom, offset_x, offset_y)

            if state != prev_state:
                buffer.erase()

                if is_global:
                    log(f"[\033[92mDEBUG\033[0m] angle_x = {angle_x}, angle_y = {angle_y}, angle_z = {angle_z}")
                    projected = project_globe(
                        globe_points,
                        angle_x, angle_y, angle_z,
                        zoom, offset_x, offset_y,
                        aspect_ratio=aspect_ratio
                    )

                else:
                    projected = project_map(
                        map_data,
                        angle_x, angle_y, angle_z,
                        zoom, offset_x, offset_y
                    )

                render_map(buffer, projected)

                lon = lat = None

                if not is_global:
                    # display lat/lon of center
                    try:
                        # multiply by stride
                        lon, lat = xy(transform, (height // 2) * 16,
                                      (width // 2) * 16)
                    except:
                        pass

                elif is_global:
                    from earthscii.globe_tile_manager import latlon_from_vector
                    lat, lon = latlon_from_vector(forward_vec)

            render_overlay(buffer, angle_x, angle_y, angle_z, zoom, lat, lon)

            buffer.noutrefresh()
            curses.doupdate()
            prev_state = state

        except KeyboardInterrupt:
            break

        time.sleep(0.016)


if __name__ == '__main__':
    try:
        curses.wrapper(main)
    except Exception as e:
        import traceback
        log(f"[\033[91mFATAL\033[0m] Uncaught Exception: {e}")
        with open("debug.log", "a") as f:
            traceback.print_exc(file=f)
        raise

def handle_keys(key, angle_x, angle_y, angle_z, zoom, offset_x, offset_y):
    changed = False

    if key == ord('q'):
        log("[\033[32mINFO\033[0m] Exiting\n------------")
        raise KeyboardInterrupt()
    elif key == ord('w'):  # tilt up
        angle_x -= 5
        changed = True
    elif key == ord('s'):  # tilt down
        angle_x += 5
        changed = True
    elif key == ord('a'):  # rotate left (yaw)
        angle_z -= 5
        changed = True
    elif key == ord('d'):  # rotate right (yaw)
        angle_z += 5
        changed = True
    elif key == ord(','):
        angle_y -= 5  # orbit left
        changed = True
    elif key == ord('.'):
        angle_y += 5  # orbit right
        changed = True
    elif key == ord('+') or key == ord('='):
        zoom *= 1.1
        changed = True
    elif key == ord('-'):
        zoom /= 1.1
        changed = True
    elif key == curses.KEY_UP:
        offset_y -= 1
        changed = True
    elif key == curses.KEY_DOWN:
        offset_y += 1
        changed = True
    elif key == curses.KEY_LEFT:
        offset_x -= 1
        changed = True
    elif key == curses.KEY_RIGHT:
        offset_x += 1
        changed = True
    elif key == ord('r'):
        angle_x, angle_y, angle_z = 0, 90, 0
        changed = True

    angle_x = max(min(angle_x, 95), -95)
    return angle_x, angle_y, angle_z, zoom, offset_x, offset_y, changed


def render_overlay(buffer, angle_x, angle_y, angle_z, zoom, lat=None, lon=None):
    buffer.addstr(0, 0, "@")  # This should always appear in top-left
    if lat is not None and lon is not None:
        buffer.addstr(0, 1, f"Lat: {lat:.4f}, Lon: {lon:.4f}")
    buffer.addstr(0, 50, f"angle_x = {angle_x}", curses.color_pair(3))
    buffer.addstr(1, 50, f"angle_y = {angle_y}", curses.color_pair(3))
    buffer.addstr(2, 50, f"angle_z = {angle_z}", curses.color_pair(3))
    buffer.addstr(3, 50, f"zoom = {zoom:.2f}", curses.color_pair(3))


def init_curses(stdscr):
    curses.curs_set(0)
    stdscr.nodelay(True)
    stdscr.timeout(50)
    stdscr.keypad(True)
    curses.start_color()

    # Define color pairs (foreground, background)
    curses.init_pair(1, curses.COLOR_BLUE, curses.COLOR_BLACK)
    curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)
    curses.init_pair(3, curses.COLOR_WHITE, curses.COLOR_BLACK)


def forward_from_angles(angle_x, angle_y):
    fx = np.cos(np.radians(angle_y)) * np.cos(np.radians(angle_x))
    fy = np.sin(np.radians(angle_x))
    fz = np.sin(np.radians(angle_y)) * np.cos(np.radians(angle_x))
    return np.array([fx, fy, fz])



import os
import requests
from pathlib import Path

TILE_DIR = Path("./tiles")
TILE_DIR.mkdir(exist_ok=True)

def tile_filename(lat, lon):
    ns = 'n' if lat >= 0 else 's'
    ew = 'e' if lon >= 0 else 'w'
    return f"{ns}{abs(lat):02d}_{ew}{abs(lon):03d}_1arc_v3.tif"

def local_tile_path(lat, lon):
    return TILE_DIR / tile_filename(lat, lon)

def download_tile(lat, lon):
    """Attempts to download a DEM tile from NASA Earthdata or other source."""
    path = local_tile_path(lat, lon)
    if path.exists():
        return str(path)

    # Example source: this is a placeholder; replace with real download URL
    filename = tile_filename(lat, lon)
    url = f"https://your-tile-server.org/tiles/{filename}"

    print(f"Downloading {filename} from {url}...")
    try:
        response = requests.get(url)
        if response.status_code == 200:
            with open(path, "wb") as f:
                f.write(response.content)
            print(f"Saved to {path}")
            return str(path)
        else:
            print(f"Failed to download tile: {response.status_code}")
            return None
    except Exception as e:
        print(f"Error downloading tile: {e}")
        return None



import rasterio
import math

EARTH_RADIUS = 6371000  # meters


def latlon_to_xyz(lat, lon, elevation):
    lat_rad = math.radians(lat)
    lon_rad = math.radians(lon)
    r = EARTH_RADIUS + elevation
    x = r * math.cos(lat_rad) * math.cos(lon_rad)
    y = r * math.cos(lat_rad) * math.sin(lon_rad)
    z = r * math.sin(lat_rad)
    return (x, y, z)


def load_etopo_as_sphere_points(path, stride=32, z_scale=1.0):
    """
    Convert a GeoTIFF file to a list of 3D sphere points.

    Args
    ----
    path: path to GeoTIFF file
    stride: sampling stride in pixels
    z_scale: scale factor for exaggerating elevation

    Returns
    -------
    points: list of 3D sphere points

    """
    points = []

    with rasterio.open(path) as dataset:
        elevation = dataset.read(1)
        transform = dataset.transform
        nodata = dataset.nodata
        height, width = elevation.shape

        for y in range(0, height, stride):
            for x in range(0, width, stride):
                z = elevation[y][x]
                if z == nodata:
                    continue
                lon, lat = rasterio.transform.xy(transform, y, x)
                p = latlon_to_xyz(lat, lon, z * z_scale)
                points.append(p)

    return points



#projection.py
"""Basic isometric projection with Y-axis rotation."""
import math
import numpy as np


def rotate_xyz(x, y, z, angle_x, angle_y, angle_z):
    """Apply 3D rotation around X, Y, then Z."""
    # Convert to radians
    ax = math.radians(angle_x)
    ay = math.radians(angle_y)
    az = math.radians(angle_z)

    # Rotate around X
    y, z = y * math.cos(ax) - z * math.sin(ax), y * math.sin(ax) + \
        z * math.cos(ax)

    # Rotate around Y
    x, z = x * math.cos(ay) + z * math.sin(ay), -x * math.sin(ay) + \
        z * math.cos(ay)

    # Rotate around Z
    x, y = x * math.cos(az) - y * math.sin(az), x * math.sin(az) + \
        y * math.cos(az)

    return x, y, z


def project_map(map_data, angle_x=0, angle_y=0, angle_z=0, zoom=1.0, offset_x=0, offset_y=0):
    projected = []
    ASPECT_RATIO = 0.5  # or tune to match your font

    # Convert angles to radians
    ax = math.radians(angle_x)
    ay = math.radians(angle_y)
    az = math.radians(angle_z)

    # Compute forward vector (orbit camera in spherical coords)
    fx = math.cos(ay) * math.cos(ax)
    fy = math.sin(ax)
    fz = math.sin(ay) * math.cos(ax)
    forward = np.array([fx, fy, fz])
    forward /= np.linalg.norm(forward)

    # Up vector (world up)
    world_up = np.array([0, 1, 0])

    # Right = up × forward
    right = np.cross(world_up, forward)
    right /= np.linalg.norm(right)

    # Recomputed up = forward × right
    up = np.cross(forward, right)

    # Get center of map
    all_points = [p for row in map_data for p in row]
    cx = sum(p[0] for p in all_points) / len(all_points)
    cy = sum(p[1] for p in all_points) / len(all_points)
    cz = sum(p[2] for p in all_points) / len(all_points)
    center = np.array([cx, cy, cz])

    for row in map_data:
        for x, y, z in row:
            point = np.array([x, y, z]) - center

            # Project point onto screen axes
            sx = int(np.dot(point, right) * zoom) + offset_x
            sy = int(np.dot(point, up) * zoom * ASPECT_RATIO) + offset_y
            sz = np.dot(point, forward)

            # Apply post-projection Z rotation to screen coords
            if angle_z != 0:
                theta = math.radians(angle_z)
                cos_z = math.cos(theta)
                sin_z = math.sin(theta)
                dx = sx - offset_x
                dy = sy - offset_y
                sx = int(dx * cos_z - dy * sin_z + offset_x)
                sy = int(dx * sin_z + dy * cos_z + offset_y)

            projected.append((sx, sy, z))

    return projected



"""Map depth to ASCII and place chars."""
import curses
from earthscii.utils import log


def render_map(buffer, projected_points):
    chars = ".,:-=+*#%@"
    if not projected_points:
        log("[\033[93mWARN\033[0m] No projected points")
        return

    height, width = buffer.getmaxyx()
    min_depth = min(p[2] for p in projected_points)
    max_depth = max(p[2] for p in projected_points)
    depth_range = max_depth - min_depth or 1
    log(f"[\033[94mRENDER\033[0m] Screen size = {width}x{height}")
    log(f"[\033[94mRENDER\033[0m] Depth range: {min_depth} to {max_depth}")
    log(f"[\033[94mRENDER\033[0m] First few points: {projected_points[:5]}")

    for x, y, depth in projected_points:
        ix, iy = int(x), int(y)
        if 0 <= iy < height and 0 <= ix < width:
            norm = (depth - min_depth) / depth_range  # normalized depth
            ch = chars[int(norm * (len(chars) - 1))]

            # Assign color based on height
            if norm < 0.045:
                color = curses.color_pair(1)
            elif norm < 0.6:
                color = curses.color_pair(2)
            else:
                color = curses.color_pair(3)

            try:
                buffer.addch(iy, ix, ch, color)
            except:
                pass # silently ignore out-of-bounds

        # Draw center marker
        try:
            buffer.addstr(height // 2, width // 2, "X",
                          curses.color_pair(3))
        except:
            log("[ERROR] Failed to draw center marker")



"""Load a DEM file and convert it to a list of 3D points."""
import rasterio


def load_tile_at(lat, lon):
    fname = f"data/n{lat:02d}_w{abs(lon):03d}_1arc_v3.tif"
    return load_dem_as_points(fname)


def load_dem_as_points(path, stride=16, z_scale=0.5, transformable=True):
    """
    Load a DEM file (GeoTIFF) and convert it to a list of 3D points.

    stride: downsample factor for performance
    """
    with rasterio.open(path) as dataset:
        elevation = dataset.read(1)
        transform = dataset.transform
        width = elevation.shape[1]
        height = elevation.shape[0]

        # mask out nodata values
        valid = elevation != dataset.nodata

        points = []
        for y in range(0, height, stride):
            row = []
            for x in range(0, width, stride):
                z = elevation[y][x]
                if z == dataset.nodata:
                    continue
                row.append((x, y, z * z_scale))
            if row:
                points.append(row)

        return (points, transform) if transformable else points



import math
import numpy as np
from earthscii.utils import log


def project_globe(points, angle_x=0, angle_y=0, angle_z=0, zoom=1.0,
                  offset_x=0, offset_y=0, screen_width=80, screen_height=24,
                  aspect_ratio=0.5):
    log(f"[\033[32mINFO\033[0m] Projecting {len(points)} points")
    projected = []
    EARTH_RADIUS = 6371000  # in meters

    # Compute camera rotation
    ax = math.radians(angle_x)
    ay = math.radians(angle_y)
    az = math.radians(angle_z)

    # Camera forward vector
    fx = math.cos(ay) * math.cos(ax)
    fy = math.sin(ax)
    fz = math.sin(ay) * math.cos(ax)
    forward = np.array([fx, fy, fz])
    forward /= np.linalg.norm(forward)

    world_up = np.array([0, 1, 0])
    right = np.cross(world_up, forward)
    right /= np.linalg.norm(right)
    up = np.cross(forward, right)

    for p in points:
        rel = np.array(p)

        # Discard back-facing hemisphere
        if np.dot(forward, rel / np.linalg.norm(rel)) < 0:
            continue

        # Dynamic scale: map full screen width to ~2 Earth radii
        SCALE = (screen_width / (2 * EARTH_RADIUS)) / zoom

        sx = int(np.dot(rel, right)* SCALE) + offset_x
        sy = int(np.dot(rel, up) * SCALE * aspect_ratio) + offset_y
        sz = np.dot(rel, forward)
        projected.append((sx, sy, sz))
        if len(projected) < 5:
            log(f"[\033[96mPOINT\033[0m] sx={sx}, sy={sy}, sz={sz:.2f}")


    log(f"[\033[32mINFO\033[0m] Projected {len(projected)} front-facing points")
    return projected



import rasterio
import math

EARTH_RADIUS = 6371000  # meters

def latlon_to_xyz(lat, lon, elevation):
    lat_rad = math.radians(lat)
    lon_rad = math.radians(lon)
    radius = EARTH_RADIUS + elevation
    x = radius * math.cos(lat_rad) * math.cos(lon_rad)
    y = radius * math.cos(lat_rad) * math.sin(lon_rad)
    z = radius * math.sin(lat_rad)
    return (x, y, z)


def load_tiles_as_sphere_points(paths, stride=16, z_scale=0.5):
    points = []

    for path in paths:
        with rasterio.open(path) as dataset:
            elevation = dataset.read(1)
            transform = dataset.transform
            width = elevation.shape[1]
            height = elevation.shape[0]
            nodata = dataset.nodata

            for y in range(0, height, stride):
                for x in range(0, width, stride):
                    z = elevation[y][x]
                    if z == nodata:
                        continue

                    lat, lon = rasterio.transform.xy(transform, y, x)
                    z_scaled = z * z_scale
                    p = latlon_to_xyz(lat, lon, z_scaled)
                    points.append(p)

    return points
